;; Generated package description from web-server.el  -*- no-byte-compile: t -*-
(define-package "web-server" "0.1.2.0.20210811.22503" "Emacs Web Server" '((emacs "24.1") (cl-lib "0.6")) :commit "5dbb1f956e7efe2065e66e3d95fe74c8b8697b86" :authors '(("Eric Schulte" . "schulte.eric@gmail.com")) :maintainer '("Eric Schulte" . "schulte.eric@gmail.com") :keywords '("http" "server" "network") :url "https://github.com/eschulte/emacs-web-server")
